
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// Written by Hans de Nivelle, September 2021. 

#ifndef PROLOG_FUNCPRED_
#define PROLOG_FUNCPRED_  1

#include <unordered_map> 
#include <unordered_set>
#include <string>
#include <iostream>

namespace prolog
{

   struct stringarity
   {
      std::string str;
      size_t ar;

      stringarity( const std::string& str, size_t ar )
         : str( str ), 
           ar( ar ) 
      { } 

      struct hash
      {
         std::hash< std::string > h;

         size_t operator( ) ( stringarity f ) const
         {
            return h( f. str ) + 11 * f. ar;
         }
      };

      struct equal_to
      {
         bool operator( ) ( stringarity f1, stringarity f2 ) const
         {
            return f1. ar == f2. ar && f1. str == f2. str;
         }
      };

   };

   inline std::ostream& 
   operator << ( std::ostream& out, stringarity f )
   {
      out << f. str << '/' << f. ar;
      return out;
   }

  
   class function 
   {

      using indextype = 
         std::unordered_set< stringarity, 
                             stringarity::hash, stringarity::equal_to > ;
   
      indextype::const_iterator p;

      function( ) = delete;

public:
      function( const std::string& str, size_t ar ) 
      { 
         static indextype index;
         p = index. insert( stringarity( str, ar )). first; 
      }

      function( const char* c, size_t ar )
         : function( std::string(c), ar )
      { }

      const std::string& getstring( ) const { return ( p -> str ); }
      size_t getarity( ) const { return ( p -> ar ); }

      bool operator == ( function f ) const { return p == f.p; } 
      bool operator != ( function f ) const { return p != f.p; } 

      struct hash
      {
         std::hash< const stringarity* > h;

         size_t operator( ) ( function f ) const 
            { return h( & *(f.p) ); }
      };

      struct equal_to
      {
         bool operator( ) ( function f1, function f2 ) const
            { return f1. p == f2. p; }
      };

      using set =
      std::unordered_set< function, function::hash, function::equal_to > ;

      template< typename D >
      using map =
      std::unordered_map< function, D, function::hash, function::equal_to > ;

   };


   inline std::ostream& operator << ( std::ostream& out, function f )
   {
      out << f. getstring( ) << "/" << f. getarity( ); 
      return out;
   }

}

#endif 



