
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// An opdef is a Prolog-style operator definition, see for example in 
//    Sterling and Shapiro, the art of Prolog. 

#ifndef PROLOG_OPDEF_
#define PROLOG_OPDEF_  1

#include <iostream>
#include <string>

namespace prolog
{

   enum optag { fx, fy, xfx, xfy, yfx, xf, yf }; 
      // x must be weaker, y can have same strength.
      // xfy = right associative.
      // yfx = left associative.
      // xfx = non-associative.
      // The general rule is that 'x' wins from 'y' when the
      // priorities are the same. That means that in
      // fx A yf, the postfix operator gets the A, so it will be
      // fx ( A yf ). 

   static const char* getcstring( optag ); 
   
   inline std::ostream& operator << ( std::ostream& out, optag tag )
   {
      out << getcstring( tag ); return out;
   }
   
   struct opdef
   {
      std::string str;
      optag type; 
      unsigned int attraction;
         // smaller number means more attraction.

      static constexpr unsigned int supattraction = 1000000;


      opdef( const std::string& str, optag type, unsigned int attraction )
         : str( str ),
           type( type ),
           attraction( attraction )
      {
         if( attraction >= supattraction )
            throw std::runtime_error( "opdef: too unattractive" ); 
      }

      bool isprefix( ) const 
         { return type == fx || type == fy; } 
      bool isinfix( ) const 
         { return type == xfx || type == xfy || type == yfx; } 
      bool ispostfix( ) const 
         { return type == xf || type == yf; } 

      // Decide how a conflict between this opdef
      // and the other opdef will be resolved. 
      // The question is how to parse T1 this T2 other T3.
      // A return value of -1 means ( T1 this T2 ) other T3.
      // A return value of 1 means T1 this ( T2 other T3 ).
      // A return value of 0 means that the conflict is a syntax error.

      int decide( const opdef& other ) const;

      bool sameside( const opdef& other ) const; 
         // True if we are on the same side, that is
         // we are both prefix, infix, or postfix operators. 
 
   };

   bool operator == ( const opdef& def1, const opdef& def2 );
   bool operator != ( const opdef& def1, const opdef& def2 );

   std::ostream& operator << ( std::ostream&, const opdef& ); 
}

#endif



