
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.




// Written by Ravinder Indarapu and Hans de Nivelle in March 2003.
// Replaced containers::hashmap by std::map in August 2005. 

#include "syntax.h"
#include <limits>

namespace prolog 
{


   syntax::syntax( ) :
      printlistsyntax( true ),
      maxprintdepth( std::numeric_limits< unsigned int > :: max( )),
      terminator( "." ) 
   { }


   bool syntax::insert( const opdef& def )
   {
      auto r = operators. equal_range( def. str ); 
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. sameside( def ))
         {
            if( p -> second != def )
            {
               p -> second = def; 
               return true;
            }
            else
               return false;    // already present. 
         }
      }
      
      operators. insert( std::pair( def. str, def ));
         // We could try to use r. second as hint, but it seems
         // not worth the effort. 

      return true; 
   }

 
   bool syntax::remove( const opdef& def )
   {
      auto r = operators. equal_range( def. str ); 
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second == def )
         {
            operators. erase(p);
            return true;
         }
      }
      return false; 
   }


   bool prolog::syntax::hasprefixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. isprefix( ))
            return true;
      }
      return false;
   }


   bool prolog::syntax::hasinfixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. isinfix( ))
            return true;
      }
      return false;
   }


   bool prolog::syntax::haspostfixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. ispostfix( ))
            return true;
      }
      return false;
   }


   const prolog::opdef& 
   prolog::syntax::prefixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. isprefix( ))
            return p -> second; 
      }
      throw std::logic_error( "prefix not found" ); 
   }


   const prolog::opdef& 
   prolog::syntax::infixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. isinfix( ))
            return p -> second; 
      }
      throw std::logic_error( "infix not found" ); 
   }

   
   const prolog::opdef& 
   prolog::syntax::postfixdef( const std::string& str ) const
   {
      auto r = operators. equal_range( str );
      for( auto p = r. first; p != r. second; ++ p )
      {
         if( p -> second. ispostfix( ))
            return p -> second; 
      }
      throw std::logic_error( "postfix not found" ); 
   }


   void 
   prolog::syntax::print( std::ostream& os ) const 
   {
      os << "Prolog Syntax:\n\n";

      os << "printlistsyntax =     " << printlistsyntax << "\n";
      os << "maxprintdepth =       " << maxprintdepth << "\n";
      os << "terminator =          " << terminator << "\n";
      os << "\n";

      os << "defined operators: \n";
      for( const auto& def : operators ) 
         os << "   " << def. second << "\n"; 
   }

} 



