
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// Written by Ravinder Indarapu and Hans de Nivelle in March 2003.
// Replaced containers::hashmap by std::map in August 2005.
// Somewhat modified in September 2006, I don't remember how. 
// Rewritten in September 2021 to work with Maphoon, and with
// std::unordered_multimap.  

#ifndef PROLOG_SYNTAX_
#define PROLOG_SYNTAX_   1

#include <unordered_map>
#include "opdef.h" 

// A syntax determines in which format terms are read and printed.
// Its main component are the definitions of Prolog-style 
// prefix/infix/postfix operator definitions.
// In addition, it defines:
//
//  - Whether list syntax is used when printing a term. List syntax is 
//    always accepted as input.
//
// - Up to which depth terms are printed. This does not affect the 
//   depth of terms that can be read. 
//
// - The string that is used for terminating input. Reasonable choices are
//   . or ; 

namespace prolog 
{

   class syntax
   {
      std::unordered_multimap< std::string, opdef > operators; 

   public: 
      bool printlistsyntax;
         //  If true, then list notation is used when printing output.
         //  List notation is always accepted as input.

      unsigned int maxprintdepth;
         //  The maximal depth up to which terms are printed. A subtree that
         //  occurs at a depth deeper than maxprintdepth, is printed as
         //  <...> . One could argue that this is not part of the syntax,
         //  but I don't know a better place to put it.

      std::string terminator; 

      syntax( );
         // Create a new empty syntax. 
         // The syntax is initially empty, the default nil is [], and the
         // default cons is the dot. By default, printlistsyntax = true. The
         // default value for maxprintdepth is +infinity. 

      bool insert( const opdef& def );
         // We insert def if either there was no definition for def.ident
         // at the given position (prefix,infix or postfix), or 
         // a conflicting definition, which we will overwrite. 
         // We return true if the insertion took place. 

      bool remove( const opdef& def );
         // Delete def, if it occurs. Otherwise do nothing.
         // Return true if the removal took place.

      bool hasprefixdef( const std::string& ident ) const;
      bool hasinfixdef( const std::string& ident ) const;
      bool haspostfixdef( const std::string& ident ) const;

      const opdef& prefixdef( const std::string& ident ) const;
      const opdef& infixdef( const std::string& ident ) const;
      const opdef& postfixdef( const std::string& ident ) const;
         // Must be checked in advance that it exists.

      void clear( ) { operators. clear( ); }

      void print( std::ostream& out ) const; 
   };


   inline std::ostream& operator << ( std::ostream& out, const syntax& synt )
   {
      synt. print( out );
      return out;
   }

}

#endif


